/*
 *  $Id: Connection.java,v 1.4 1996/01/19 01:51:54 iansmith Exp $
 *  $Author: iansmith $
 *  $Log: Connection.java,v $
 * Revision 1.4  1996/01/19  01:51:54  iansmith
 * Documentation improvements in the javadoc comments.
 *
 * Revision 1.3  1996/01/15  18:45:10  iansmith
 * Fixed a bug in the selectOrExamine code in which it was always making
 * mailboxes read-only. I fixed it to correctly look at the protocol
 * response to make this choice.
 *
 * Revision 1.2  1996/01/08  05:52:05  iansmith
 * Added throws InternalErr to all the major API functions and some
 * internal functions.  This was required because some lower level code is
 * now throwing it and I need to pass it through.
 *
 *  Revision 1.1  1996/01/06 19:49:13  iansmith
 *  Initial revision
 *
 */
package GIT.GVU.IMAP;
import java.lang.*;
import java.util.*;

/**
 * Class to hold onto a logged-in connection to an IMAP server. 
 * This class is generally used to get at mailboxes (SELECT them). This
 * class is basically a representation of the Authenticated state in 
 * RFC1730.  Once you have used SELECT or EXAMINE, you are in the 
 * selected state of the protocol and then call getMailbox() to get the
 * mailbox object that has your mail data.
 *
 * @version $Id: Connection.java,v 1.4 1996/01/19 01:51:54 iansmith Exp $
 * @author Ian Smith
 */

public class Connection {
  /**
   * Constructor for connection. Requires a protocol argument because we
   * need a link back to our network.
   * 
   * @param p protocol object whic owns this connectio
   */
  Connection(Protocol p) {
    protocol=p;
    mailbox=null;
    mailboxNames=null;
  }
  /**
   * This protocol object is our link back to the network.
   */
  Protocol protocol;

 /** 
   * Forward protocol dispatching 
   * 
   * @param s the string to dispatch 
   * @exception InternalErr is propagated
   * @return true if we dispatched the string or someone downstream did, otherwise false
   */
  boolean walkForward(String s) throws InternalErr {

    /* try to dispatch it locally...*/
    if (Util.lookingAt(s,"* LIST")==true) {
      handleList(s);
      return true;
    }
    /* nobody wanted it... pass around with other objects */
    /* ask our mailbox if he wants it */
    if (mailbox!=null) {
      /* he's there, did he handle it? */
      if (mailbox.walkForward(s)==true) {
	/* he dealt with it */
	return true;
      }
    }
    return false;
  }
  /** 
   * Backward protocol dispatching 
   * 
   * @param s the string to dispatch  
   * @exception InternalErr is propagated
   * @return true if we dispatched the string or someone downstream did, otherwise false
   */
  boolean walkBackward(String s) throws InternalErr {

    /* first, check to see if our mailbox wants it... do we hava mailbox? */
    if (mailbox!=null) {
      /* he consumed it */
      if (mailbox.walkBackward(s)==true) {
	return true;
      }
    }
    /* no issue with the mailbox object, so just handle it locally */
    
    /* end of local processing */
    return false;
  }
  /**
   * Select a mailbox for use on this connection. 
   * 
   * @param s name of the mailbox
   * @exception IMAPException is propagated
   * @exception InternalErr is propagated
   * @return true if the command was successful, false otherwise
   */
  public boolean SELECT(String s) throws InternalErr, IMAPException {
    return (selectOrExamine(true,s));
  }
  /** 
   * Examine a mailbox (i.e. select but readonly) on this connection.
   *
   * @param s name of the mailbox
   * @exception IMAPException is propagated
   * @exception InternalErr is propagated
   * @return true if the command succeeded, false otherwise
   */
  public boolean EXAMINE(String s) throws InternalErr, IMAPException {
    return (selectOrExamine(false,s));
  }
  /**
   * This is an internal use function that invokes either SELECT or
   * EXAMINE depending on its first parameter. It then dispatches the
   * resulting protocol messages and returns the true if the command 
   * succeeded.
   *
   * @param doSelect boolean which is true if want select, false for examine
   * @param s mailbox name to select or examine
   * @exception IMAPException is propagated
   * @exception InternalErr is propagated
   * @return true if command succeeded, false otherwise
   */
  boolean selectOrExamine(boolean doSelect,String s) throws InternalErr, IMAPException {
    Vector result=new Vector();
    String cmd,last;
    boolean readonly=true;

    /* which command is it? Select?*/
    if (doSelect==true) {
      cmd=new String("SELECT");
    } else {
      cmd=new String("EXAMINE");
    }
    /* send protocol message */
    protocol.getIO().roundTrip(cmd+" "+s,result);
    /* check to see if we succeeded */
    if (Util.replyToken(result,"OK")==true) {
      /* we are cool */
      /* figure out if we have a read-write or read-only mailbox */
      /* by looking at the last line of the response */
      last=(String)result.elementAt(result.size()-1);
      /* look for the string READ-WRITE in there */
      if (last.indexOf("READ-WRITE")!=-1) {
	readonly=false;
      }
      /* make the new mailbox */
      mailbox=new Mailbox(readonly,this);
      /* send the messages to be dispatched ... we do this *after* we*/
      /* create the mailbox object because he is going to be handling */
      /* a lot of them */
      protocol.dispatchVector(result,false);
      /* tell the user he is cool */
      return true;
    }
    /* send the messages to be dispatched ... there might be messages in*/
    /* there that we were not expecting */
    protocol.dispatchVector(result,false);
    /* after a failed examine, mailbox is cleared */
    mailbox=null;
    /* failed to examine */
    return false;
  }
  /** 
   * List the available mailboxes on this connection.  Note: you can always
   * list mailboxes, even when selected.<P>
   * 
   * To list the mailboxes in the user foo's Mail directory use:
   * <LIST>EXAMINE("~foo/Mail/","*");</LIST><P>
   * 
   * @param refname string with reference name (prefix)
   * @param mailbox string with mailbox name (suffix)
   * @exception IMAPException is propagated
   * @exception InternalErr is propagated
   * @return true if LIST succeeded, false otherwise
   */
  public boolean LIST(String refname, String mailbox) throws InternalErr, IMAPException {
    Vector result=new Vector(100,50); /* could be big if a home directory */

    /* clear out the list of mailbox names */
    mailboxNames=new Vector(100,500);
    /* send the command to the IMAP server */
    protocol.getIO().roundTrip("LIST " + refname + " " + mailbox,result);
    /* dispatch results */
    protocol.dispatchVector(result,false);
    /* tell caller if we made it ok */
    return (Util.replyToken(result,"OK"));
    
  }
  /** This variable holds the current mailbox 
   */
  Mailbox mailbox;
  /** This function returns the current mailbox. Note: you can always call
   * this function but it will return null if no mailbox is selected. Also,
   * if you call SELECT or EXAMINE when you are already selecting a mailbox
   * the value of this function WILL change. It will replace the old mailbox
   * with the new mailbox or set it to null if the SELECT or EXAMINE fails.
   */
  public Mailbox getMailbox() { return mailbox;};
  /**
   * This function processes responses to the LIST command. 
   * 
   * @param s string with the server response string 
   */
  void handleList(String s) {
    StringTokenizer st=new StringTokenizer(s);
    String name;
    String flags;
    Vector v;
    boolean selectable, interesting;
    MailboxName mname;

    /* skip over the * */
    st.nextToken();
    /* now the LIST */
    st.nextToken();
    /* get the list of flags */
    flags=st.nextToken(")");
    /* get rid of the now in place right paren and space */
    st.nextToken(" ");
    /* now the delimiter field XXX WE EXPECT UNIX AND / XXX */
    st.nextToken();
    /* now grab the name */
    name=st.nextToken();

    /* put the list of flags into v */
    v=Util.parseStringList(flags);
    /* figure out if the object is selectable or not */
    /*
     * I'm not sure if this is really right or not, but I am assuming that
     * unless it is marked selectable it is selectable. This is not really
     * clear from RFC1730 as there is this "\Noinferiors" foolishness. It
     * seems to be just a marker for the boolean opposite of selectable,
     * but I'm not really sure.
     */
    selectable=(Util.stringInVector(v,"\\NoSelect")==false);
    /*
     * now figure out if they are interesting... again it doesn't seem 
     * like we should need both \Marked and \Unmarked, but we have them.
     *
     * This whole function may be XXX based on these assumptions.
     */
    interesting=(Util.stringInVector(v,"\\Marked")==true);
    /* make a new mailbox name based on what we learned*/
    mname=new MailboxName(name,selectable,interesting);
    /* add this new guy to the list of names */
    mailboxNames.addElement(mname);
    }
  /**
   * This function returns a vector of MailboxNames (although you'll have
   * to cast them to get them out. This is a the set of MailboxName objects
   * which was created with the last LIST command. Note: If you call LIST
   * with bad parameters or with a mailbox specifiers which don't result
   * in any mailbox names you'll get an empty vector from this function.
   *
   * @return a vector containing zero or more mailbox objects
   */
  public Vector getMailboxNames() { return mailboxNames;};
  /**
   * This variable holds the mailbox names for this connection.
   */
  Vector mailboxNames;
  /**
   * This function creates a new mailbox. You'll need to be careful to
   * use the correct separator character if you want to create a directory
   * on the server side. 
   * 
   * You can not and should not CREATE an INBOX. It is always there.
   * 
   * @s string name of the new mailbox 
   * @exception IMAPException is propagated
   * @exception InternalErr is propagated
   * @return true if the creation succeeded, false otherwise
   */
  public boolean CREATE(String s) throws InternalErr, IMAPException {
    return(simpleCommand("CREATE",s));
  }
  /** 
   * This is an internal use function which centralizes the code for
   * simple commands. These are commands which just use the command 
   * name and one parameter. These commands then just return true or 
   * false to the caller.
   * 
   * @param c string with the command name
   * @param p string with the parameter
   * @exception IMAPException is propagated
   * @exception InternalErr is propagated
   * @return true if command succeeded, false otherwise
   */
  boolean simpleCommand(String c,String p) throws InternalErr, IMAPException{
    /* where to put results */
    Vector result=new Vector();
    /* send it to the server */
    protocol.getIO().roundTrip(c + " " + p, result);
    /* dispatch the results */
    protocol.dispatchVector(result,false);
    /* check to see if we succeeded */
    if (Util.replyToken(result,"OK")==true) {
      return true;
    }
    /* failed */
    return false;
  }
  /**
   * This command <B>PERMANENTLY</B> removes a mailbox. Be careful!
   *
   * You can not and should not DELETE the INBOX. The system needs it.
   *
   * @param s string with the name of the mailbox to delete
   * @exception IMAPException is propagated
   * @exception InternalErr is propagated
   * @return true only if the mailbox was actually deleted
   */
  public boolean DELETE(String s) throws InternalErr, IMAPException {
    return (simpleCommand("DELETE",s));
  }
  /** 
   * This command renames a mailbox. Note: As per RFC 1730, you can rename
   * the INBOX. If you do so, a new INBOX is created. Note: You cannot 
   * DELETE or CREATE the INBOX so this seems a little funny to me.
   *
   * @param s string with the mailbox's current name
   * @param n string with the mailbox's new name
   * @exception IMAPException is propagated
   * @exception InternalErr is propagated
   * @return true if the RENAME succeeded, false otherwise
   */
  public boolean RENAME(String s, String n) throws InternalErr, IMAPException{
    /* where to put results */
    Vector result=new Vector();
    /* send it to the server */
    protocol.getIO().roundTrip("RENAME " + s + " " + n,result);
    /* dispatch the results */
    protocol.dispatchVector(result,false);
    /* check to see if we succeeded */
    if (Util.replyToken(result,"OK")==true) {
      return true;
    }
    /* failed */
    return false;
  }
  /**
   * This function is for internal use when a mailbox has been closed. It updates
   * the Connection's internal state to reflect this.
   */
  void mailboxClosed() {
    mailbox=null;
  }
  /**
   * This function returns the protocol object associated with this connection. 
   * 
   * @return the Protcol object of this Connection object
   */
  Protocol getProtocol() { return protocol;}
}
