/*
 *  $Id: Protocol.java,v 1.4 1996/01/19 01:58:02 iansmith Exp $
 *  $Author: iansmith $
 *  $Log: Protocol.java,v $
 * Revision 1.4  1996/01/19  01:58:02  iansmith
 * Documentation improvements in the javadoc comments.
 *
 * Added support for an Asynch.Action object to be used in conjunction
 * with the BYE message. So I added a function to set it and modified
 * handleBye to deal with the AA.
 *
 * Revision 1.3  1996/01/10  22:01:40  iansmith
 * Currently I'm not debugging the graveyard.
 *
 * Revision 1.2  1996/01/08  06:06:16  iansmith
 * RCS keyword cleanup.
 *
 *  Revision 1.1  1996/01/06 19:50:57  iansmith
 *  Initial revision
 *
 */

/*
 * This is the lowest level of the java IMAP system. It creates network 
 * connections to a server. In its initial state, it can do nothing. You
 * connect it to a server with connect().  Once connected, you then
 * use the LOGIN and LOGOUT messages to control the state of the connection.
 * If you are logged-in (successful call of LOGIN()) you can call 
 * getConnection() to get the current connection object. 
 *
 */

package GIT.GVU.IMAP;

/* Need all the java language of course */
import java.lang.*;
import java.util.*;
import java.io.*;

/**
 * Class to initiate and control the protocol level exchange with an
 * IMAP server. This class is used *before* connection is established as
 * well as to hold onto a non-authenticated connection before logging in.
 * See RFC1730 for more info on the system states.
 *
 * @version $Id: Protocol.java,v 1.4 1996/01/19 01:58:02 iansmith Exp $
 * @author Ian Smith
 */
public class Protocol {
  /*
   * Need to keep track of internal state of our object 
   */
  boolean connected;
  Connection connection; /* XXX admittedly, this is a misnomer XXX */
  /* 
   * Do we want to print out the messages that end up in the graveyard
   */
  static final boolean debugGraveyard=false;
  /* actual IO object which talks to the server */
  IO io;

  /** 
   * Create a protocol object. This can't do anything useful until you
   * call connect. 
   *
   * @see #connect
   */
  public Protocol() {
    /* initialize internal state */
    connected=false;
    connection=null;
    io=null;
  }
  /**
   * Connect to server.  Stupidly, this doesn't result in a connection
   * object, the LOGIN() command does.
   * 
   * @param s the String with the server name to connect to
   * @exception IMAPException is thrown if the connection cannot be completed (check the detail message)
   * @exception InternalErr is thrown if we fail to ascertain the capability of the IMAP server
   */
  public void connect(String s) throws IMAPException, InternalErr {
    Vector result=new Vector(); /* place to hold the result of request */

    /* can't connect if we are already connected */
    if (connected==true) {
      throw new IMAPException("Can't connect: already connected!");
    }
    /* go ahead and try to create the IO object */
    try {
      io=new IO(s);
    } catch (IOException e) {
      throw new IMAPException("Unable to connect to IMAP server: " + e);
    }
    /* ok, got a connection */
    connected=true;
    /* just for verification... send a capability command */
    io.roundTrip("CAPABILITY",result);
    /* check for bogusly formed capability command */
    if (Util.replyToken(result,"OK")==false) {
      throw new InternalErr("Badly formed CAPABILITY request or CAPABILITY request denied");
    }
    /* dispatch the capability response, and maybe other stuff */
    dispatchVector(result,false);
  }
  /**
   * Login with a connection to the IMAP server. This will allow you to
   * proceed and call getConnection(). Note: BOGUS! This passes your 
   * password over the network in the clear! We are no worse off than 
   * rlogin or telnet I guess, but I should be using AUTHENTICATE 
   * instead of LOGIN.
   *
   * @param u the user name for whom this authentication is being checked
   * @param p the password of this user
   * @exception IMAPException if the protocol state is violated
   * @exception InternalErr is thrown if we get a message back from the server that the LOGIN was badly formed... we do this do make it easy to distinguish bad logins from bad protocol since bad protocol is very unlikely.
q   * @return true if the login procedure succeeded, false if not
   */
  public boolean LOGIN(String u, String p) throws InternalErr,IMAPException {
    Vector result=new Vector();

   /* do a sanity check */
    if ((connected==false) || (connection!=null)) {
      throw new IMAPException("Unable to perform LOGIN when not connected " +
                              "to a server or if already logged in to a " +
                              "server");
    }
    /* ok, send a protocol request */
    io.roundTrip("LOGIN " + u + " " + p,result);
    /* change the state of the system if any messages came in */
    dispatchVector(result,false);
    /* check for ok */
    if (Util.replyToken(result,"OK")==false) {
      /* uh oh, bad login or badly formed protocol ?*/
      if (Util.replyToken(result,"BAD")==true) {
	throw new InternalErr("Badly formed LOGIN request");
      }
      /* just typed the wrong password */
      return false;
    }
    /* make a connection object with us as its protocol */
    connection=new Connection(this);
    /* everything when ok */
    return true;
  }
  /** 
   * Dispatch a vector of protocol input strings. 
   *
   * @param v the vector of strings
   * @param b whether or not to dispatch the final message in the vector 
   * @exception InternalErr
   */
  void dispatchVector(Vector v, boolean b) throws InternalErr {
    int i;

    /* loop over all elemenents */
    for (i=0; i<v.size(); ++i) {
      /* if we are on the last one, check to see if we should dispatch it */
      if (i==v.size()-1) {
	/* did they want it dispatched */
	if (b==false) {
	  /* nope */
	  continue;
	}
      }
      /* normal case: dispatch the current element of the vector */
      dispatch((String)v.elementAt(i));
    }
  }
  /** 
   * Forward protocol dispatching 
   * 
   * @param s the string to dispatch 
   * @exception InternalErr
   * @return true if we dispatched the string or someone downstream did, otherwise false
   */
  boolean walkForward(String s) throws InternalErr {

    /* try to dispatch it locally...*/
    /* CAPABILITY */
    if (Util.lookingAt(s,"* CAPABILITY")==true) {
      handleCapability(s);
      return true;
    }
    /* BYE */
    if (Util.lookingAt(s,"* BYE")==true) {
      handleBye(s);
      return true;
    }
    /* nobody wanted it... pass around with other objects */
    /* do we have a connection yet? */
    if (connection==null) {
      /* no connection, so nobody else to possibly want it*/
      return false;
    }
    /* ask our connection if he wants it */
    return connection.walkForward(s);
  }
  /** 
   * Backward protocol dispatching 
   * 
   * @param s the string to dispatch 
   * @exception InternalErr
   * @return true if we dispatched the string or someone downstream did, otherwise false
   */
  boolean walkBackward(String s) throws InternalErr {

    /* first, check to see if our connection wants it */
    if (connection!=null) {
      /* we have a connection, does he want it? */
      if (connection.walkBackward(s)==true) {
	/* he's handled it, just blow out of here */
	return true;
      }
    }
    /* no issue with the connection object, so just handle it locally */
    
    /* end of local processing */
    return false;
  }

  /**
   * Dispatch one protocol input string
   *
   * @param s the string to dispatch 
   * @exception InternalErr
   */
  void dispatch(String s) throws InternalErr {
    /* The dispatch process consists of two phases.  */
    /* Phase 1: A walk "backward" across the string of protocol objects */
    /*          to see if anyone wants to handle the string. This should */
    /*          only be used by objects which need to avoid the normal */
    /*          processing order.*/
    /* Phase 2: A walk "forward" across the string of protocol objects */
    /*          This is the normal route for dispatching the messages */
    
    /* A backward walk goes Mailbox->Connection->Protocol whereas a  */
    /* foward walk goes Protcol->Connection->Mailbox */
    
    /* check for someone wanting to snarf in backward walk */
    if (walkBackward(s)==false) {
      /* nobody grabbed it, try a normal walk */
      if (walkForward(s)==false) {
	/* nobody processed it all */
	graveyard(s);
      }
    }
  }
  /**
   * This function is called when no processing is done on an IMAP message. 
   *
   * @param s string that no other part of the system processed
   */
  void graveyard(String s) {
    /* we are printing debugging messages, then dump this one out */
    if (debugGraveyard==true) {
      System.out.println("GRAVEYARD: " + s);
    }
  }
  /** 
   * Method to handle the server's untagged response to a CAPABILITY request.
   * As per the IMAP v4 spec, we just ignore everything in the response other
   * than IMAP4 since that is all we use.
   * 
   * @param s string which was the text of the response 
   * @exception InternalErr
   */
  void handleCapability(String s) throws InternalErr {
    StringTokenizer st=new StringTokenizer(s);
    String tok;
    /* walk over all tokens looking for the IMAP4 signifier*/
    while (st.hasMoreElements()) {
      /* get the next token */
      tok=st.nextToken();
      /* see if its IMAP4 */
      if (tok.equals("IMAP4")==true) {
	/* it, is so we are done */
	return;
      }
    }
    /* if we get here, there was no IMAP4, so raise an error */
    throw new InternalErr("This system was designed for IMAP v4 and we " +
                          "don't appear to be talking to an IMAP 4 server.");
  }
  /**
   * Logout of an IMAP server. This will both disconnect and log you out, so
   * further connections will require a connect() and a LOGIN().
   * 
   * @exception IMAPException if the protocol state is violated
   * @exception InternalErr  is propagated
   * @return true if logout was successful, false otherwise
   */
  public boolean LOGOUT() throws IMAPException, InternalErr {
    Vector result=new Vector();

    /* do a sanity check */
    if (connected==false) {
      throw new IMAPException("Unable to perform LOGOUT when not connected " +
                              "to a server");
    }
    
    /* send the message and wait for response */
    io.roundTrip("LOGOUT",result);
    /* deal with the BYE in his response (possibly) */
    dispatchVector(result,false);
    /* determine if we succeeded or not */
    return (Util.replyToken(result,"OK")==true);
  }
  /** This is called to handle a BYE message from the server. This does
   * the cleanup necessary to get ready for a new connection in the future. 
   *
   * @param s string with the BYE message. It gets ignored. 
   */
  void handleBye(String s) {
    int index;
    /* just get rid of the connection... the GC will fix it! */
    connection=null;
    /* again, just blow away the IO object... I assume it has a finalizer */
    /* which will disconnect the network when it dies */
    io=null;
    /* we are not going to be connected either */
    connected=false;
    /* check for action object */
    if (byeAction!=null) {
      /* where is the BYE message */
      index=s.indexOf("BYE");
      /* everything after the BYE is the reason ... invoke the action*/
      byeAction.run("BYE",s.substring(index+3));
    }
  }
  /** This command is used to "do nothing" to the server. It returns true or
   * false to tell if it worked or not. It is useful in that it resets the
   * autologout timer inside the server and may result in the number 
   * messages changing in a mailbox (due to new mail).  Also, the status of
   * messages in mailboxes might have changed and might be updated.
   *
   * Note: this call doesn't require you have called LOGIN() yet, 
   * but it isn't terribly useful if you haven't.
   * 
   * @exception IMAPException is thrown if the protocol state is violated
   * @exception InternalErr is propagated
   * @return true if NOOP was successful, false otherwise 
   */
  public boolean NOOP() throws IMAPException, InternalErr {
    Vector result=new Vector();

    /* do a sanity check */
    if (connected==false) {
      throw new IMAPException("Unable to perform NOOP when not connected");
    }
    /* send it to the server */
    io.roundTrip("NOOP",result);
    /* deal with the results  */
    dispatchVector(result,false);
    /* return true if we got an OK */
    return (Util.replyToken(result,"OK")==true);
  }
  /** 
   * Get the connection object associated with this protocol object. Note:
   * you can call this anytime, but you'll get a null if we are not yet
   * logged in and connected and all that mess.
   *
   * @return current connection object
   */
  public Connection getConnection() { return connection;};
  /**
   * Get the IO object associated with this connection. 
   *
   * @return current IO object 
   */
  public IO getIO() { return io;};
  /**
   * This object holds the asynchaction object which is to be notified if we get 
   * the BYE message. 
   */
  AsynchronousAction byeAction;
  /**
   * This lets you set the object that gets called if we get the BYE message from
   * the server. By the time the action object gets notified,  the Connection object
   * associated with this Protocol object (and everything further downstream) is
   * already gone. You should probably use this to inform the user what is happening.
   * The parameter given to this Action object is a String with the reason (if any)
   * for the demise of this connection.  
   * 
   * @param AsynchronousAction a object to be invoked on a BYE message
   */
  public void setByeAction(AsynchronousAction a) { byeAction=a;};
}
